/* Sentinel - IRC Statistical and Operator Services
** datadump.c - Raw data dump function
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: datadump.c,v 1.15 2003/04/07 14:32:48 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"

void datadump(char *filename)
{
  FILE *fp = fopen(filename, "w");
  time_t now = time(NULL);
  struct Server *s;
  int i;

  if (!fp)
  {
    log(L_CRIT, "Unable to open data dump file %s (%s)", filename,
        strerror(errno));
    return ;
  }

  fprintf(fp, "NETINFO %u %ld %ld %ld %d %d %ld %d %d %ld %d %d "
          "%ld %d %ld %ld %ld %ld\n",
          netinfo.away, netinfo.users, netinfo.maxusers,
          (long)netinfo.t_maxusers, netinfo.opers, netinfo.maxopers,
          (long)netinfo.t_maxopers, netinfo.servers, netinfo.maxservers,
          (long)netinfo.t_maxservers,
#ifdef IRCD_2_10
          netinfo.services, netinfo.maxservices,
          (long)netinfo.t_maxservices, netinfo.maxaway,
          (long)netinfo.t_maxaway,
#elif defined(IRCD_HYB7) || defined(IRCD_HYBSVS) || defined(IRCD_BAHAMUT)
          0, 0, 0L, netinfo.maxaway, (long)netinfo.t_maxaway,
#else
          0, 0, 0L, 0, 0L,
#endif
          netinfo.channels, netinfo.maxchannels,
          (long)netinfo.t_maxchannels);

  fprintf(fp, "DAILY %ld %ld %ld %d %ld %d %ld %d %ld %ld %ld\n",
          (long)now,
          daily.users, (long)daily.t_users, daily.opers,
          (long)daily.t_opers, daily.servers, (long)daily.t_servers,
#ifdef IRCD_2_10
          daily.services, (long)daily.t_services,
#else
          0, 0L,
#endif
          daily.channels, (long)daily.t_channels);

  for (i = 0; i < S_TABLE_SIZE; i++)
  {
    for (s = serverlist[i]; s; s = s->next)
    {
      fprintf(fp,
              "%s %d %ld %ld %ld %d %d %ld %d %u %u %d %d %ld "
              "%d %d %ld\n",
              s->name, IsSHide(s) ? 1 : 0, s->users,
              s->stats->maxusers, (long)s->stats->t_maxusers,
              s->opers, s->stats->maxopers,
              (long)s->stats->t_maxopers, s->stats->numsplits,
              s->stats->operkills, s->stats->serverkills,
#ifdef IRCD_2_10
              s->services, s->stats->maxservices,
              (long)s->stats->t_maxservices, s->away,
              s->stats->maxaway, (long)s->stats->t_maxaway);
#elif defined(IRCD_HYB7) || defined(IRCD_BAHAMUT) || defined(IRCD_HYBSVS)

              0, 0, 0L, s->away, s->stats->maxaway,
              (long)s->stats->t_maxaway);
#else
              0, 0, 0L, 0, 0, 0L);
#endif

    }
  }

  fclose(fp);
}
