/* Sentinel - IRC Statistical and Operator Services
** dlink.c - Hybrid derived doubly linked list library
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: dlink.c,v 1.10 2003/04/07 14:32:49 wcampbel Exp $
*/

/* Original header */

/*  tcm-hybrid/src/tools.c by fl_
 *  Copyright (C) 2002 ircd-hybrid development team
 *
 *  #Id: tools.c,v 1.6 2002/06/24 16:21:51 leeh Exp #
 */

#include <stdlib.h>
#include "stats.h"
#include "dlink.h"

int dlink_count = 0;

dlink_node *
dlink_create(void)
{
  dlink_node *m;

  m = (dlink_node *) malloc(sizeof(dlink_node));
  m->data = NULL;
  m->next = NULL;
  m->prev = NULL;

  dlink_count++;

  return m;
}

/* XXX - macro? */
void
dlink_free(dlink_node *m)
{
  dlink_count--;
  free(m);
}

void
dlink_add(void *data, dlink_node *m, dlink_list *list)
{
  m->data = data;
  m->next = list->head;
  m->prev = NULL;

  if (list->head != NULL)
    list->head->prev = m;
  else if (list->tail == NULL)
    list->tail = m;

  list->head = m;
  list->count++;
}

void
dlink_add_before(dlink_node * b, void *data, dlink_node * m, dlink_list * list)
{
  /* Shortcut - if its the first one, call dlinkAdd only */
  if (b == list->head)
  {
    dlink_add(data, m, list);
  }
  else
  {
    m->data = data;
    b->prev->next = m;
    m->prev = b->prev;
    b->prev = m;
    m->next = b;
    list->count++;
  }
}

void
dlink_add_tail(void *data, dlink_node *m, dlink_list *list)
{
  m->data = data;
  m->next = NULL;
  m->prev = list->tail;

  if (list->head == NULL)
    list->head = m;
  else if (list->tail != NULL)
    list->tail->next = m;

  list->tail = m;
  list->count++;
}

void
dlink_delete(dlink_node *m, dlink_list *list)
{
  /* item is at head */
  if (m->prev == NULL)
    list->head = m->next;
  else
    m->prev->next = m->next;

  /* item is at tail */
  if (m->next == NULL)
    list->tail = m->prev;
  else
    m->next->prev = m->prev;

  list->count--;
}

dlink_node *
dlink_find(void *data, dlink_list *list)
{
  dlink_node *ptr;

  DLINK_FOREACH(ptr, list->head)
  {
    if (ptr->data == data)
      return ptr;
  }

  return NULL;
}

dlink_node *
dlink_find_delete(void *data, dlink_list *list)
{
  dlink_node *ptr;

  DLINK_FOREACH(ptr, list->head)
  {
    if (ptr->data != data)
      continue;

    if (ptr->next != NULL)
      ptr->next->prev = ptr->prev;
    else
      list->tail = ptr->prev;

    if (ptr->prev != NULL)
      ptr->prev->next = ptr->next;
    else
      list->head = ptr->next;

    ptr->next = ptr->prev = NULL;

    list->count--;

    return ptr;
  }

  return NULL;
}

int
dlink_length(dlink_list *list)
{
  return list->count;
}

