/* Sentinel - IRC Statistical and Operator Services
** services.c - Functions to track IRCnet services
**
** Copyright W. Campbell and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: services.c,v 1.12 2003/04/07 14:32:49 wcampbel Exp $
*/

#include "stats.h"
#include "struct.h"
#include "proto.h"
#include "match.h"

#ifdef IRCD_2_10
struct Service *findservice(char *nick, char *server)
{
  struct Service *s = NULL;
  dlink_node *dl;

  debug("findservice(%s, %s)", nick, server);

  DLINK_FOREACH(dl, servicehead.head)
  {
    s = (struct Service *) dl->data;
    /* XXX - Do I need to match() the server? */
    if (!(irccasecmp(nick, s->nick)) &&
        !(strcasecmp(server, s->server)))
      return s;
  }

  return NULL;
}

void AddService(char *nick, char *server, char *dist, char *gecos)
{
  struct Service *s = smalloc(sizeof(struct Service));
  dlink_node *dl = dlink_create();
  struct Server *sv = findserver(server);
  time_t current = time(NULL);

  strlcpy(s->nick, nick, MAXNICK + 1);
  strlcpy(s->server, server, MAXHOST);
  strlcpy(s->dist, dist, MAXHOST);
  strlcpy(s->gecos, gecos, LINE);
  dlink_add(s, dl, &servicehead);
  netinfo.services++;
  if (netinfo.services > netinfo.maxservices)
  {
    netinfo.maxservices = netinfo.services;
    netinfo.t_maxservices = current;
  }
  if (netinfo.services > daily.services)
  {
    daily.services = netinfo.services;
    daily.t_services = current;
  }
  if (sv)
  {
    sv->services++;
    if (sv->services > sv->stats->maxservices)
    {
      sv->stats->maxservices = sv->services;
      sv->stats->t_maxservices = current;
    }
  }
}

void DelService(char *nick, char *server)
{
  struct Service *s = NULL;
  struct Server *sv = findserver(server);
  dlink_node *dl;

  DLINK_FOREACH(dl, servicehead.head)
  {
    s = (struct Service *) dl->data;
    if (!(irccasecmp(nick, s->nick)) &&
        !(strcasecmp(server, s->server)))
    {
      dlink_delete(dl, &servicehead);
      free(s);
      dlink_free(dl);
      netinfo.services--;
      if (sv)
        sv->services--;
      return ;
    }
  }

  log(L_WARNING, "Unable to delete service %s@%s", nick, server);
}

void DelServicesByServer(char *server)
{
  struct Service *s = NULL;
  dlink_node *dl;
  dlink_node *tdl;

  DLINK_FOREACH_SAFE(dl, tdl, servicehead.head)
  {
    s = (struct Service *) dl->data;
    if (!(strcasecmp(server, s->server)))
    {
      dlink_delete(dl, &servicehead);
      dlink_free(dl);
      free(s);
      netinfo.services--;
      /* We're getting rid of this server record...
      ** not necessary to decrease its services count
      */
    }
  }
}

#endif

